searchNodes=[{"doc":"Caveat emptor: Protean is likely to undergo significant changes and should not be considered stable. A library for managing state and side-effects with event-driven statecharts. Protean was initially inspired by XState , a robust JavaScript/TypeScript statechart implementation, but strays to adhere to Elixir and OTP conventions. We also follow much of the SCXML W3C Standard 's recommendations, but compatibility is not a goal. What are statecharts? They are an extension to finite state machines that allow you to model complex behavior in a declarative, data-driven manner. They include nested and parallel states, enhanced/augmented state (through assigns), side-effects (through actions), process management (through spawn), and more. To learn more about statecharts, I recommend statecharts.dev . Goals This project is currently an exploration of statecharts as they fit into the assigns of Elixir and OTP. XState adopted the actor model in its implementation, so Elixir seemed like a natural fit. However, it may be that Elixir/OTP makes these abstractions unnecessary. Example This simple statechart has a single state that defines the behavior of a counter with an optional maximum and minimum. defmodule Counter do use Protean alias Protean.Action @machine [ initial : :active , assigns : [ count : 0 , min : nil , max : nil ] , states : [ atomic ( :active , on : [ match ( &quot;Inc&quot; , actions : :increment , guard : [ not : :at_max ] ) , match ( &quot;Dec&quot; , actions : :decrement , guard : [ not : :at_min ] ) , match ( { &quot;Set&quot; , _ } , actions : :set_min_or_max ) , match ( { &quot;Log&quot; , _ } , actions : :log ) ] ) ] ] @impl true def handle_action ( :increment , context , _event ) do context |&gt; Action . assign_in ( [ :count ] , &amp; &amp;1 + 1 ) end def handle_action ( :decrement , context , _event ) do context |&gt; Action . assign_in ( [ :count ] , &amp; &amp;1 - 1 ) end def handle_action ( :set_min_or_max , context , { &quot;Set&quot; , { key , val } } ) do context |&gt; Action . assign ( key , val ) end def handle_action ( :log , context , { &quot;Log&quot; , attribute } ) do %{ assigns : assigns } = context IO . puts ( &quot; \#{ attribute } : \#{ assigns [ attribute ] } &quot; ) context end @impl true def guard ( :at_max , %{ assigns : %{ max : max , count : count } } , _event ) do max &amp;&amp; count &gt;= max end def guard ( :at_min , %{ assigns : %{ min : min , count : count } } , _event ) do min &amp;&amp; count &lt;= min end end It can be started under a supervisor, but we'll start it directly using Protean's built-in DynamicSupervisor . { :ok , pid } = Protean . start_machine ( Counter ) Protean . current ( pid ) . assigns # %{count: 0, min: nil, max: nil} Protean . send ( pid , &quot;Inc&quot; ) # :ok Enum . each ( 1 .. 4 , fn _ -&gt; Protean . send ( pid , &quot;Inc&quot; ) end ) Protean . current ( pid ) . assigns # %{count: 5, min: nil, max: nil} Protean . call ( pid , { &quot;Set&quot; , { :max , 10 } } ) # %Protean.Context{ # assigns: %{count: 5, max: 10, min: nil}, # event: {&quot;Set&quot;, {:max, 10}}, # value: MapSet.new([[&quot;active&quot;, &quot;#&quot;]]) # } Enum . each ( 1 .. 20 , fn _ -&gt; Protean . send ( pid , &quot;Inc&quot; ) end ) Protean . send ( pid , { &quot;Log&quot; , :count } ) # count: 10 Defining a statechart Protean machines are event-driven statecharts , which means that, unlike ordinary finite-state machines, they can have complex, nested, potentially parallel states. This is more easily visualized than read, and I highly recommend looking at XContext's introduction to state machines and statecharts for that reason. Refer to Protean.Builder for documentation on machine definitions. When use Protean is invoked, functions and macros from Protean.Builder are imported automatically. The @machine attribute By default, Protean assumes that your machine is defined on the @machine attribute. defmodule MyMachine do use Protean @machine [ initial : :my_initial_state , states : [ atomic ( :my_initial_state , # ... ) , # ... ] ] end Starting supervised machines Since state machines typically model structured interactions with a defined beginning and end, they will generally be started under a DynamicSupervisor . Protean starts one (as well as a Registry ) by default, in order to manage subprocesses that are started during machine execution through the use of Protean.Builder.proc/2 et al. Machines can be started under this supervisor using Protean.start_machine/2 . { :ok , machine } = Protean . start_machine ( MyMachine ) Similar to GenServer , calling use Protean will also define a child_spec/1 that allows you to start a machine in a standard supervision tree, if you wish: children = [ Counter ] Supervisor . start_link ( children , strategy : :one_for_one ) Interacting with Protean machines Under the hood, a Protean machine is a GenServer , and Protean exposes a similar set of functions for interacting with one. You can see the individual docs for the functions in this module for details on their behavior, but here are some highlights. Familiar functions call/3 - Send an event synchronously to a Protean machine and receive the machine context and any replies resulting from transition. send/2 - Send an event asynchronously to a Protean machine. Always returns :ok . send_after/3 - Send an event to a Protean machine after a given delay. Like Process.send_after/4 , returns a timer reference so that the send can be canceled with Process.cancel_timer/2 . Additional functions specific to Protean machines current/1 - Get the current machine context of a running Protean machine. matches?/2 - Query the currently active state(s) of a machine. subscribe/2 (and unsubscribe/2 ) - Subscribes the calling process to receive a message on every state transition. Subscriptions You can subscribe to a Protean machine to receive messages when the machine transitions. This functionality depends on the optional dependency :phoenix_pubsub . To use it, add the following to deps in your mix.exs : defp deps do [ :phoenix_pubsub # ... ] end If you are already starting a Phoenix.PubSub in your application (e.g. a Phoenix application), you need to configure the :protean application to use your process instead of starting its own. This can be done by adding the following to your config.exs : config :protean , :pubsub , name : MyApp.PubSub , start : false For subscription usage, see subscribe/2 .","ref":"Protean.html","title":"Protean","type":"behaviour"},{"doc":"Makes a synchronous call to the machine, awaiting any transitions that result. Returns a tuple of {context, replies} , where context is the next state of the machine, and replies is a (possibly empty) list of replies returned by action callbacks resulting from the event.","ref":"Protean.html#call/3","title":"Protean.call/3","type":"function"},{"doc":"Synchronously retrieve the current machine context. TODO: Allow optional timeout as with call/3 .","ref":"Protean.html#current/1","title":"Protean.current/1","type":"function"},{"doc":"Optional callback for defining dynamic delays. Example @machine [ # ... states : [ will_transition : [ after : [ delay : &quot;my_delay&quot; , target : &quot;new_state&quot; ] ] , new_state : [ # ... ] ] ] @impl true def delay ( &quot;my_delay&quot; , context , _ ) do context . assigns [ :configured_delay ] || 1000 end","ref":"Protean.html#c:delay/3","title":"Protean.delay/3","type":"callback"},{"doc":"Optional callback to determine whether a conditional transition should occur. Example @machine [ # ... states : [ editing_user : [ on : [ { { :user_commit , _ } , guard : :valid_user? , actions : [ &quot;broadcast&quot; ] , target : &quot;viewing_user&quot; } , { { :user_commit , _ } , guard : { :not , :valid_user? } , actions : [ &quot;show_invalid_user_error&quot; ] } ] ] ] ] @impl true def guard ( :valid_user? , context , { _ , user } ) do User . changeset ( % User { } , user ) . valid? end","ref":"Protean.html#c:guard/3","title":"Protean.guard/3","type":"callback"},{"doc":"Optional callback for actions specified in response to a transition. Receives the current machine context and event triggering the action as arguments. Returns one of: context - same as {:noreply, context} {:noreply, context} - the machine context with any new actions {:reply, reply, context} - a reply and the machine context with any new actions Example @machine [ # ... on : [ match ( { :data , _any } , target : :data_received , actions : [ :assign_data , :broadcast_data ] ) ] ] @impl true def handle_action ( :assign_data , context , { :data , data } ) do context |&gt; Protean.Action . assign ( :last_received , data ) end def handle_action ( :broadcast_data , context , _ ) do %{ notify : pid , last_received : data } = context . assigns PubSub . broadcast! ( @pubsub , @topic , data ) context = context |&gt; Protean.Action . send ( { :data , data } , to : pid ) { :reply , data , context } end","ref":"Protean.html#c:handle_action/3","title":"Protean.handle_action/3","type":"callback"},{"doc":"Returns true if the machine is currently in the given state. Note that calling matches?/2 on a machine process is a synchronous operation that is equivalent to: machine |&gt; Protean . current ( ) |&gt; Protean . matches? ( descriptor )","ref":"Protean.html#matches?/2","title":"Protean.matches?/2","type":"function"},{"doc":"Sends an asynchronous event to the machine. Shares semantics with GenServer.cast/2 .","ref":"Protean.html#send/2","title":"Protean.send/2","type":"function"},{"doc":"Sends an event to the machine after time in milliseconds has passed. Returns a timer reference that can be canceled with Process.cancel_timer/1 .","ref":"Protean.html#send_after/3","title":"Protean.send_after/3","type":"function"},{"doc":"Optional callback for spawned processes linked to a certain machine state. Should return a child_spec appropriate for the type of process being spawned: Protean.Builder.proc/2 Protean.Builder.task/2 Protean.Builder.stream/2 Example @machine [ states : [ atomic ( :awaiting_task , spawn : [ task ( :my_task , done : :completed ) ] ) , atomic ( :completed ) ] ] @impl true def spawn ( :task , :my_task , context , _event ) do { __MODULE__ , :run_task , [ context . assigns . task_data ] } end","ref":"Protean.html#c:spawn/4","title":"Protean.spawn/4","type":"callback"},{"doc":"Start a Protean machine linked to Protean's supervisor. By default, machines will be registered and named using Protean's process management registry. Options :assigns - assigns map that will be merged into the default machine context. :machine - defaults to the machine defined in module - machine configuration. :module - defaults to module - callback module used for actions, guards, spawns, etc. See &quot;Callbacks&quot;. :parent - defaults to self() - process id of the parent that will receive events from the machine if a Protean.Action.send(..., to: :parent) action is used or when the machine reaches a state with :type of :final . Any option accepted by GenServer.start_link/3 .","ref":"Protean.html#start_machine/2","title":"Protean.start_machine/2","type":"function"},{"doc":"TODO","ref":"Protean.html#stop/3","title":"Protean.stop/3","type":"function"},{"doc":"Subscribes the caller to receive messages when a machine transitions. Returns {:ok, id} where id is a unique identifier for the machine process, or {:error, error} if subscription is unsuccessful. Note: Subscriptions depend on Phoenix.Pubsub , an optional dependency. Options :filter - if set to :replies , the caller will only be sent messages with replies. Examples { :ok , id } = Protean . subscribe ( machine ) Protean . send ( machine , :some_event ) # receive: {^id, context, []} You can also subscribe to only receive messages if replies are non-empty: { :ok , id } = Protean . subscribe ( machine , filter : :replies ) Protean . send ( machine , :reply_triggering_event ) # receive: {^id, context, [reply, ...]}","ref":"Protean.html#subscribe/2","title":"Protean.subscribe/2","type":"function"},{"doc":"Unsubscribes the caller from machine transition messages.","ref":"Protean.html#unsubscribe/1","title":"Protean.unsubscribe/1","type":"function"},{"doc":"Any message sent to a Protean machine.","ref":"Protean.html#t:event/0","title":"Protean.event/0","type":"type"},{"doc":"Unique identifier for a Protean machine process.","ref":"Protean.html#t:id/0","title":"Protean.id/0","type":"type"},{"doc":"A running Protean machine process.","ref":"Protean.html#t:machine/0","title":"Protean.machine/0","type":"type"},{"doc":"Option values for Protean machines.","ref":"Protean.html#t:machine_option/0","title":"Protean.machine_option/0","type":"type"},{"doc":"Return values of start_machine/2","ref":"Protean.html#t:on_start/0","title":"Protean.on_start/0","type":"type"},{"doc":"","ref":"Protean.html#t:spawn_type/0","title":"Protean.spawn_type/0","type":"type"},{"doc":"Option values for start* functions.","ref":"Protean.html#t:start_option/0","title":"Protean.start_option/0","type":"type"},{"doc":"Option values for use Protean .","ref":"Protean.html#t:using_option/0","title":"Protean.using_option/0","type":"type"},{"doc":"TODO","ref":"Protean.Interpreter.Features.html","title":"Protean.Interpreter.Features","type":"module"},{"doc":"","ref":"Protean.Interpreter.Features.html#install/1","title":"Protean.Interpreter.Features.install/1","type":"function"},{"doc":"Interpreter execution hooks. The purpose of this module is to define extensions at various points of machine interpretation. See the description of the interpretation loop in Protean.Interpreter before reading on. While hooks can implement &quot;standalone&quot; behaviors, such as ignoring a certain type of event, they are used by Protean to implement higher-level features such as :spawn . See Protean.Interpreter.Features for more. Hooks event_filter is used to filter out or modify events. after_event_filter is used to perform setup that you'd only want to occur if the event was not filtered out. after_microstep runs after a microstep has occurred. after_macrostep runs after a macrostep has occurred. on_stop runs when interpretation stops. See function documentation for the individual hooks below.","ref":"Protean.Interpreter.Hooks.html","title":"Protean.Interpreter.Hooks","type":"module"},{"doc":"Register an after_event_fitler hook. Function must accept two arguments, interpreter and event , and must return {:cont, interpreter }.","ref":"Protean.Interpreter.Hooks.html#after_event_filter/2","title":"Protean.Interpreter.Hooks.after_event_filter/2","type":"function"},{"doc":"Register an after_macrostep hook. Function must accept an interpreter and return {:cont, interpreter }.","ref":"Protean.Interpreter.Hooks.html#after_macrostep/2","title":"Protean.Interpreter.Hooks.after_macrostep/2","type":"function"},{"doc":"Register an after_microstep hook. Function must accept interpreter and return {:cont, interpreter }.","ref":"Protean.Interpreter.Hooks.html#after_microstep/2","title":"Protean.Interpreter.Hooks.after_microstep/2","type":"function"},{"doc":"Register an event_filter hook. Function must accept one argument, {interpreter, event} and return one of: {:cont, {interpreter, event}} - continue with interpreter and event {:halt, interpreter} - halt with interpreter and ignore event","ref":"Protean.Interpreter.Hooks.html#event_filter/2","title":"Protean.Interpreter.Hooks.event_filter/2","type":"function"},{"doc":"Register an on_stop hook. Function must accept an interpreter and return {:cont, interpreter }.","ref":"Protean.Interpreter.Hooks.html#on_stop/2","title":"Protean.Interpreter.Hooks.on_stop/2","type":"function"},{"doc":"Protean manages state, processes, and side-effects through actions , data structures describing things that should occur as a result of a transition. Introduction When a Protean machine transitions from one state to the next (or even self-transitions back to the same state), the interpreter collects any actions that should be performed as a result of that transition. Actions are then executed in a specific order: Exit actions - Any :exit actions specified on states that are exiting as a result of the transition. Transition actions - Any :actions specified on the transition that responded to the event. Entry actions - Any :entry actions specified on the states being entered. Actions are a unifying abstraction that serve many purposes: assign to and update a machine's context; produce additional actions to be run immediately after; run arbitrary side-effects, like broadcasting a PubSub message; reply to the sender of the event with a value; spawn additional processes whose lifecycle can be tied to machine execution. I recommend reading the sections below to gain a better understanding of how and where actions can be applied, then review the &quot;Callback Actions&quot; and &quot;Inline Actions&quot; provided by this module. Handling an action The most common way to use actions is through the Protean.handle_action/3 callback. This callback is run when an action is specified like this: [ # ... entry : [ :entering_state ] on : [ match ( % MyEvent { } , actions : [ :first_action , :second_action ] ) ] , exit : [ :exiting_state ] ] These are then handled in callbacks: @impl true def handle_action ( :first_action , context , % MyEvent { data : data } ) do # ... { :reply , reply , context } end See Protean.handle_action/3 for possible return values and their effect. Action callbacks must always return the machine context, but they can attach actions to that context that will be immediately executed by the interpreter. For instance: def handle_action ( :update_data , context , { :data_updated , changes } ) do context |&gt; Action . update_in ( [ :data ] , &amp; Map . merge ( &amp;1 , changes ) ) end In this case, update_in/3 is being used to update some data in the machine :assigns . But, we could perform additional actions if we wish, such as: def handle_action ( :update_data , context , { :data_updated , changes } ) do %{ topic : topic , other_process : pid , data : data } = context . assigns new_data = Map . merge ( data , changes ) PubSub . broadcast! ( @pubsub , topic , { :data_updated , changes } ) context |&gt; Action . send ( { :data_commited , new_data } , to : pid ) |&gt; Action . assign ( :data , new_data ) end Note about return values These last two examples have returned the context directly, which is equivalent to returning {:noreply, context} . You can also emit replies that will be available to the sender of the event using {:reply, reply, context} . This can be very useful for creating client APIs around your machines, much like you would do with GenServer . The best practice is to let the interpreter execute as many actions as possible, as opposed to performing explicit side-effects in the callback. In many cases, such as spawning additional processes, this allows Protean to supervise and tie process lifecycle to the machine's lifecycle (or changes in state). Low-level API: the Action behaviour Ultimately, all actions resolve to Protean.Action.t/0 , a simple struct containing a module and an argument. The module is expected to implement the Protean.Action behaviour. Action execution, then, follows a simple logic: if the action being executed is a Protean.Action struct, call the exec_action/2 callback on the module, passing it the associated argument and the interpreter in its current state. Otherwise, wrap it in a struct that delegates to the callback module associated with the machine. This is how Protean.handle_action/3 is called. Many of features features boil down to syntax sugar over actions, including :spawn and delayed transitions using :after .","ref":"Protean.Action.html","title":"Protean.Action","type":"behaviour"},{"doc":"Same as assign/2 but used inline in machine configuration.","ref":"Protean.Action.html#assign/1","title":"Protean.Action.assign/1","type":"function"},{"doc":"Attach an action that merges the given assigns into a machine's context.","ref":"Protean.Action.html#assign/2","title":"Protean.Action.assign/2","type":"function"},{"doc":"Attach an action that assigns a value to a key in a machine's context.","ref":"Protean.Action.html#assign/3","title":"Protean.Action.assign/3","type":"function"},{"doc":"Same as assign_in/3 but used inline in machine configuration.","ref":"Protean.Action.html#assign_in/2","title":"Protean.Action.assign_in/2","type":"function"},{"doc":"Attach an action that assigns into a machine's context. Similar to Kernel.put_in/3 .","ref":"Protean.Action.html#assign_in/3","title":"Protean.Action.assign_in/3","type":"function"},{"doc":"Same as choose/2 but used inline in machine configuration.","ref":"Protean.Action.html#choose/1","title":"Protean.Action.choose/1","type":"function"},{"doc":"Attach an action that executes the first of a list of actions whose guard is truthy.","ref":"Protean.Action.html#choose/2","title":"Protean.Action.choose/2","type":"function"},{"doc":"Accepts the action_arg passed with the action as well as the Protean interpreter. Returns one of three values: {:cont, interpreter} to continue running the action pipeline {:cont, interpreter, [action]} to inject actions that will be run immediately before running the rest of the pipeline's actions {:halt, interpreter} to halt the action pipeline, canceling any further actions","ref":"Protean.Action.html#c:exec_action/2","title":"Protean.Action.exec_action/2","type":"callback"},{"doc":"Attach a custom action that implements the Protean.Action behaviour.","ref":"Protean.Action.html#put/2","title":"Protean.Action.put/2","type":"function"},{"doc":"Same as send/3 but used inline in machine configuration.","ref":"Protean.Action.html#send/2","title":"Protean.Action.send/2","type":"function"},{"doc":"Attach an action that will send a message to a process.","ref":"Protean.Action.html#send/3","title":"Protean.Action.send/3","type":"function"},{"doc":"Same as update/2 , but inserted inline in machine config.","ref":"Protean.Action.html#update/1","title":"Protean.Action.update/1","type":"function"},{"doc":"Attach an action that calls a function with a machine's context and merges the result into assigns.","ref":"Protean.Action.html#update/2","title":"Protean.Action.update/2","type":"function"},{"doc":"Same as update_in/3 but used inline in machine configuration.","ref":"Protean.Action.html#update_in/2","title":"Protean.Action.update_in/2","type":"function"},{"doc":"Attach an action that applies an update function into a machine's context. Similar to Kernel.update_in/3 .","ref":"Protean.Action.html#update_in/3","title":"Protean.Action.update_in/3","type":"function"},{"doc":"Allowed return value from exec_action/2 .","ref":"Protean.Action.html#t:exec_action_return/0","title":"Protean.Action.exec_action_return/0","type":"type"},{"doc":"Data structure representing an executable action.","ref":"Protean.Action.html#t:t/0","title":"Protean.Action.t/0","type":"type"},{"doc":"API for defining Protean machines. This module is imported by default when use Protean is spawned. Defining a machine At the outermost level, machines are specified as a keyword list, usually associated with the @machine module attribute of the defining module. @machine [ states : [ # ... ] ] For the most part, a machine definition is similar to a compound or parallel state definition, except that it allows for an addition :assigns option to specify the default assigns for the machine context. @machine [ assigns : %{ # ... } , # ... ] The top-level machine can be parallel by specifying type: :parallel : @machine [ type : :parallel , states : [ # ... ] ] See compound/2 and parallel/2 for corresponding options.","ref":"Protean.Builder.html","title":"Protean.Builder","type":"module"},{"doc":"Builds an atomic state. Atomic states are simple states that cannot define children, but represent some intermediary state of the machine. states : [ atomic ( :loading , # ... ) ] Options :spawn - list of processes to spawn, see proc/2 , task/2 , stream/2 ; :entry - actions to execute when entering this state; :exit - actions to execute when exiting this state; :always - transitions to immediately take when their guard is true, see transition/1 ; :after - transitions to take after a given delay, see delay/2 ; :on - transitions to take in response to an event, see match/2 .","ref":"Protean.Builder.html#atomic/2","title":"Protean.Builder.atomic/2","type":"function"},{"doc":"Builds a compound state. Compound states have children defined by a :states list, of which only one will be active at a given time. They additional define an :initial attribute specifying which child should become active if we transition directly to the compound state. states : [ compound ( :parent , initial : :child_a , states : [ atomic ( :child_a ) ] ) ] Compound states can define a :done transition that will be taken if one of its final children become active. In the example below, the :parent state will transition to its sibling if the final :child_b state is entered. states : [ compound ( :parent , initial : :child_a , done : transition ( target : :sibling ) , states : [ atomic ( :child_a , # ... ) , final ( :child_b ) ] ) ] Options :initial (required) - child state to enter when entering the compound state; :states (required) - one or more child states; :done - transition to take if a final child state is entered; all options available to atomic/2 .","ref":"Protean.Builder.html#compound/2","title":"Protean.Builder.compound/2","type":"function"},{"doc":"Builds a delayed transition. Delayed transitions run automatically after the given delay so long as the machine is still in the state that defined it and any given guard allows it. Accepts the same options as transition/1 .","ref":"Protean.Builder.html#delay/2","title":"Protean.Builder.delay/2","type":"function"},{"doc":"Builds a final state. Final states are a variation of atomic states that represent some form of completion. Final states cannot define transitions of their own, but entering a final state can trigger a transition in a compound or parallel parent. See compound/2 and parallel/2 . states : [ final ( :completed ) ] Options :entry - actions to execute when entering this state; :exit - actions to execute when exiting this state (as a result of a parent transition).","ref":"Protean.Builder.html#final/2","title":"Protean.Builder.final/2","type":"function"},{"doc":"Builds a pattern-matching event transition. Accepts the same options as transition/1 . Example on : [ match ( { :event_with_payload , _payload } , action : :save_payload ) , match ( % Events.OtherEvent { } , target : :other ) ]","ref":"Protean.Builder.html#match/2","title":"Protean.Builder.match/2","type":"macro"},{"doc":"Builds a parallel state. Parallel states have child states defined by a :states list, all of which will be considered active concurrently when the parallel state is active. states : [ parallel ( :parent , states : [ atomic ( :child_a , entry : :child_a_action ) , atomic ( :child_b , entry : :child_b_action ) ] ) ] In the example above, transitioning to :parent would enter both child states and cause both of their entry actions to execute. Parallel states can define a :done transition that will be taken when all of its children are in a final state. Usually, this means the parallel state's children are compound states with active final children. states : [ parallel ( :parent , done : transition ( target : :sibling ) , states : [ compound ( :compound_a , states : [ atomic ( :a_child1 ) , final ( :a_child2 ) ] ) , compound ( :compound_b , states : [ atomic ( :b_child1 ) , final ( :b_child2 ) ] ) ] ) ] In the example above, the parent parallel state will transition to its sibling once both compound states have active final children. Options :states (required) - one or more child states, all of which will be concurrently entered when the parallel state becomes active; :done - transition to take when all children are in a final state. all options available to atomic/2 .","ref":"Protean.Builder.html#parallel/2","title":"Protean.Builder.parallel/2","type":"function"},{"doc":"Specification for a spawned process linked to a parent state.","ref":"Protean.Builder.html#proc/2","title":"Protean.Builder.proc/2","type":"function"},{"doc":"Specification for a spawned stream linked to a parent state.","ref":"Protean.Builder.html#stream/2","title":"Protean.Builder.stream/2","type":"function"},{"doc":"Specification for a spawned task linked to a parent state.","ref":"Protean.Builder.html#task/2","title":"Protean.Builder.task/2","type":"function"},{"doc":"Builds a transition. Options :target - the target state of the transition; :actions - one or more actions that should be executed when the transition occurs; :guard - condition that must be true in order for the transition to occur. Guards See Protean.Guard TODO","ref":"Protean.Builder.html#transition/1","title":"Protean.Builder.transition/1","type":"function"},{"doc":"","ref":"Protean.Builder.html#t:action/0","title":"Protean.Builder.action/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:actions/0","title":"Protean.Builder.actions/0","type":"type"},{"doc":"Additional state stored in the machine context. The specified assigns will be converted to a map/0 .","ref":"Protean.Builder.html#t:assigns/0","title":"Protean.Builder.assigns/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:atomic_state_option/0","title":"Protean.Builder.atomic_state_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:atomic_state_options/0","title":"Protean.Builder.atomic_state_options/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:compound_machine_option/0","title":"Protean.Builder.compound_machine_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:compound_state_option/0","title":"Protean.Builder.compound_state_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:compound_state_options/0","title":"Protean.Builder.compound_state_options/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:delayed_transition/0","title":"Protean.Builder.delayed_transition/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:delayed_transition_option/0","title":"Protean.Builder.delayed_transition_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:delayed_transitions/0","title":"Protean.Builder.delayed_transitions/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:event_transition/0","title":"Protean.Builder.event_transition/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:event_transitions/0","title":"Protean.Builder.event_transitions/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:final_state_option/0","title":"Protean.Builder.final_state_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:final_state_options/0","title":"Protean.Builder.final_state_options/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:machine_options/0","title":"Protean.Builder.machine_options/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:parallel_machine_option/0","title":"Protean.Builder.parallel_machine_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:parallel_state_option/0","title":"Protean.Builder.parallel_state_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:parallel_state_options/0","title":"Protean.Builder.parallel_state_options/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:spawn_option/0","title":"Protean.Builder.spawn_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:spawn_options/0","title":"Protean.Builder.spawn_options/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:spawns/0","title":"Protean.Builder.spawns/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:state/0","title":"Protean.Builder.state/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:state_name/0","title":"Protean.Builder.state_name/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:states/0","title":"Protean.Builder.states/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:transition/0","title":"Protean.Builder.transition/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:transition_option/0","title":"Protean.Builder.transition_option/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:transition_options/0","title":"Protean.Builder.transition_options/0","type":"type"},{"doc":"","ref":"Protean.Builder.html#t:transitions/0","title":"Protean.Builder.transitions/0","type":"type"},{"doc":"State available to observers of and callbacks in a machine. Functions in this module should rarely be used directly. Instead, rely on the API exposed by Protean and Protean.Action to query and modify machine context.","ref":"Protean.Context.html","title":"Protean.Context","type":"module"},{"doc":"See Protean.matches?/2 .","ref":"Protean.Context.html#matches?/2","title":"Protean.Context.matches?/2","type":"function"},{"doc":"","ref":"Protean.Context.html#t:assigns/0","title":"Protean.Context.assigns/0","type":"type"},{"doc":"","ref":"Protean.Context.html#t:private_state/0","title":"Protean.Context.private_state/0","type":"opaque"},{"doc":"","ref":"Protean.Context.html#t:t/0","title":"Protean.Context.t/0","type":"type"},{"doc":"","ref":"Protean.Context.html#t:value/0","title":"Protean.Context.value/0","type":"type"},{"doc":"Protocol for guarded transitions, actions, etc. Default implementations are provided for: BitString - Call callback module with string, context, and event Atom - Call callback module with atom, context, and event Function - Call function with context and event Tuple - Higher-order guard utilities: {:and, [guard1, ...]} {:or, [guard1, ...]} {:not, guard} {:in, query} - Delegates to Protean.Context.matches?/2","ref":"Protean.Guard.html","title":"Protean.Guard","type":"protocol"},{"doc":"","ref":"Protean.Guard.html#allows?/4","title":"Protean.Guard.allows?/4","type":"function"},{"doc":"","ref":"Protean.Guard.html#t:t/0","title":"Protean.Guard.t/0","type":"type"},{"doc":"","ref":"Protean.MachineConfig.html","title":"Protean.MachineConfig","type":"module"},{"doc":"Data structure representing a Protean machine.","ref":"Protean.MachineConfig.html#__struct__/0","title":"Protean.MachineConfig.__struct__/0","type":"function"},{"doc":"Compute the full set of active nodes for the given states.","ref":"Protean.MachineConfig.html#active/2","title":"Protean.MachineConfig.active/2","type":"function"},{"doc":"Fetch a node by its id. Raises if the node cannot be found.","ref":"Protean.MachineConfig.html#fetch!/2","title":"Protean.MachineConfig.fetch!/2","type":"function"},{"doc":"Compute the initial context for a machine configuration, including any entry actions that result from entering the default states.","ref":"Protean.MachineConfig.html#initial_context/1","title":"Protean.MachineConfig.initial_context/1","type":"function"},{"doc":"Return the full lineage of the given id, including itself and all of its ancestors.","ref":"Protean.MachineConfig.html#lineage/2","title":"Protean.MachineConfig.lineage/2","type":"function"},{"doc":"","ref":"Protean.MachineConfig.html#new/2","title":"Protean.MachineConfig.new/2","type":"function"},{"doc":"User-defined machine configuration.","ref":"Protean.MachineConfig.html#t:config/0","title":"Protean.MachineConfig.config/0","type":"type"},{"doc":"Internal representation of a Protean machine","ref":"Protean.MachineConfig.html#t:t/0","title":"Protean.MachineConfig.t/0","type":"type"},{"doc":"TODO Event descriptors Guards Internal Exact Actions","ref":"Protean.Transition.html","title":"Protean.Transition","type":"module"},{"doc":"Return the actions associated with a transition","ref":"Protean.Transition.html#actions/1","title":"Protean.Transition.actions/1","type":"function"},{"doc":"","ref":"Protean.Transition.html#domain/1","title":"Protean.Transition.domain/1","type":"function"},{"doc":"Checks whether the transition is enabled for the given event.","ref":"Protean.Transition.html#enabled?/4","title":"Protean.Transition.enabled?/4","type":"function"},{"doc":"","ref":"Protean.Transition.html#new/1","title":"Protean.Transition.new/1","type":"function"},{"doc":"","ref":"Protean.Transition.html#t:t/0","title":"Protean.Transition.t/0","type":"type"},{"doc":"Execution logic for a Protean machine. Overview of interpretation loop At a high level, an SCXML-confirming statechart (which Protean is based on) defines an interpreter loop that maintains certain properties laid out in the SCXML specification. In addition to executing purely-functional state transformations, the interpreter is responsible for executing actions, side-effects, and tracking process state. The interpreter loop consists of the following steps: Check if interpreter is running; if not, exit. Execute any automatic transitions that are active in the current state, recursively until there are no more automatic transitions to be run. Process internal event if one is present. a. Microstep the interpreter by running any transitions and executing any resulting actions. b. Loop back to 1. Await an external event. Process external event in the same manner as 3. Note that the loop above contains a conceptually &quot;blocking&quot; operation in awaiting an external event. Protean.Interpreter.start/1 executes the first 3 steps of the interpreter, taking any automatic transitions that are immediately active based on the machine configuration, etc. This guarantees that the machine ends in a stable state where no more automatic transitions or internal events are left, at which point we can wait for an external event. Macrosteps and microsteps The execution of this loop until it is waiting for an external event (or interpretation has stopped) is called a macrostep . The execution of a single set of transitions resulting from an automatic transition, internal event, or external event is called a microstep . Macrosteps can consist of one or more microsteps. After this initialization phase, the trigger for the loop to continue is that we've received an external event. In the context of Protean, it's easier to think of this as initialization and event handling. Initialization: Check if the interpreter is running; if not, exit. Execute any automatic transitions that are active, looping back to 1. When there are none left, continue. Process internal event if one is present, looping back to 1. This guarantees that the machine is in a stable state with no more automatic transitions or internal events left to run. Handle external event: Run any transitions and execute actions associated with the external event. Check if the interpreter is running; if not, exit. (A transition may have caused the machine to enter a final state.) Execute any automatic transitions that have become active as a result of the event, looping back to 2. Process internal event if one is present, looping back to 2. As with initialization, this guarantees that the machine is in a stable state, ready to handle the next external event. The SCXML specification references &quot;run-to-completion&quot;, which refers to the property that, after interpretation has started, there should be no (visible) point where the interpreter is in a non-stable state. When the interpreter starts, it transitions until it is stable. When it receives an event, it transitions until it is stable.","ref":"Protean.Interpreter.html","title":"Protean.Interpreter","type":"module"},{"doc":"","ref":"Protean.Interpreter.html#add_many_internal/2","title":"Protean.Interpreter.add_many_internal/2","type":"function"},{"doc":"Handle an event, executing any transitions, actions, and side-effects associated with the current machine context. Returns a tuple of the interpreter and any replies resulting from actions that were run.","ref":"Protean.Interpreter.html#handle_event/2","title":"Protean.Interpreter.handle_event/2","type":"function"},{"doc":"Create a new Interpreter . The returned interpreter will still need to be started, which could result in additional side-effects. See start/1 .","ref":"Protean.Interpreter.html#new/1","title":"Protean.Interpreter.new/1","type":"function"},{"doc":"Whether the interpreter has been started and can accept events.","ref":"Protean.Interpreter.html#running?/1","title":"Protean.Interpreter.running?/1","type":"function"},{"doc":"Entrypoint for the interpreter that must be called before the interpreter will be in a state where it can handle external events. This is necessary in order to handle any initializing actions, spawns, or automatic transitions. Calling start/1 on an already-running interpreter is a no-op.","ref":"Protean.Interpreter.html#start/1","title":"Protean.Interpreter.start/1","type":"function"},{"doc":"Stop an interpreter, preventing further event processing.","ref":"Protean.Interpreter.html#stop/1","title":"Protean.Interpreter.stop/1","type":"function"},{"doc":"","ref":"Protean.Interpreter.html#t:t/0","title":"Protean.Interpreter.t/0","type":"type"},{"doc":"Purely-functional statechart core based on the SCXML specification. Provides the underlying state-transition logic for a statechart, primarily through transition/3 , a higher-level transition API for using %Protean.Machine{} apart from the interpreter provided by Protean, and take_transitions/3 , a lower-level API used by a statechart interpreter. It is uncommon to use this module independently of the Protean behaviour.","ref":"Protean.Machinery.html","title":"Protean.Machinery","type":"module"},{"doc":"Select any machine transitions that apply to the given event in the current context.","ref":"Protean.Machinery.html#select_transitions/4","title":"Protean.Machinery.select_transitions/4","type":"function"},{"doc":"","ref":"Protean.Machinery.html#take_transitions/3","title":"Protean.Machinery.take_transitions/3","type":"function"},{"doc":"Transition in response to an event.","ref":"Protean.Machinery.html#transition/3","title":"Protean.Machinery.transition/3","type":"function"},{"doc":"Internal representation of an individual node in a Protean.MachineConfig struct.","ref":"Protean.Node.html","title":"Protean.Node","type":"module"},{"doc":"Given a Node id, return a list containing that id and all of its ancestors.","ref":"Protean.Node.html#ancestor_ids/1","title":"Protean.Node.ancestor_ids/1","type":"function"},{"doc":"","ref":"Protean.Node.html#common_ancestor_id/1","title":"Protean.Node.common_ancestor_id/1","type":"function"},{"doc":"","ref":"Protean.Node.html#common_ancestor_id/2","title":"Protean.Node.common_ancestor_id/2","type":"function"},{"doc":"Tests whether descendant_id is in fact a descendant of ancestor_id . Returns false if they are the same node.","ref":"Protean.Node.html#descendant?/2","title":"Protean.Node.descendant?/2","type":"function"},{"doc":"Return the entry actions associated with the node.","ref":"Protean.Node.html#entry_actions/1","title":"Protean.Node.entry_actions/1","type":"function"},{"doc":"Sort the given nodes in entry order.","ref":"Protean.Node.html#entry_order/1","title":"Protean.Node.entry_order/1","type":"function"},{"doc":"Return the exit actions associated with the node.","ref":"Protean.Node.html#exit_actions/1","title":"Protean.Node.exit_actions/1","type":"function"},{"doc":"Sort the given nodes in exit order.","ref":"Protean.Node.html#exit_order/1","title":"Protean.Node.exit_order/1","type":"function"},{"doc":"","ref":"Protean.Node.html#leaf?/1","title":"Protean.Node.leaf?/1","type":"function"},{"doc":"Given a Node id, return its parent's id. Returns nil if the Node is the root.","ref":"Protean.Node.html#parent_id/1","title":"Protean.Node.parent_id/1","type":"function"},{"doc":"Resolve a Node to its leaves (atomic or final) by either returning the given node or following the node's children.","ref":"Protean.Node.html#resolve_to_leaves/1","title":"Protean.Node.resolve_to_leaves/1","type":"function"},{"doc":"An atomic node is a node without child states.","ref":"Protean.Node.html#t:atomic/0","title":"Protean.Node.atomic/0","type":"type"},{"doc":"A complex Node is one that has child nodes.","ref":"Protean.Node.html#t:complex/0","title":"Protean.Node.complex/0","type":"type"},{"doc":"A compound node is a node that defines children, of which only one can be active. It must additionally define an :initial attribute, the id of the child state that should default to active if the compound state is entered.","ref":"Protean.Node.html#t:compound/0","title":"Protean.Node.compound/0","type":"type"},{"doc":"A final node is a type of atomic node that represents some form of completion, and can therefore define no transitions itself. Note, however, that activating a final node causes an event to be dispatched that a parent node can choose to handle.","ref":"Protean.Node.html#t:final/0","title":"Protean.Node.final/0","type":"type"},{"doc":"ID encompasses the node and all its ancestors. For example, a node :child_a defined as a child of a node :parent_a might have the id [:child_a, :parent_a]","ref":"Protean.Node.html#t:id/0","title":"Protean.Node.id/0","type":"type"},{"doc":"A leaf Node is one that cannot have child states.","ref":"Protean.Node.html#t:leaf/0","title":"Protean.Node.leaf/0","type":"type"},{"doc":"","ref":"Protean.Node.html#t:nodes/0","title":"Protean.Node.nodes/0","type":"type"},{"doc":"A parallel node defines children, all of which are entered when the parallel node is entered.","ref":"Protean.Node.html#t:parallel/0","title":"Protean.Node.parallel/0","type":"type"},{"doc":"A simple Node is one that cannot have child nodes.","ref":"Protean.Node.html#t:simple/0","title":"Protean.Node.simple/0","type":"type"},{"doc":"A Node is a node in a nested state machine. See the type docs for individual nodes for more details.","ref":"Protean.Node.html#t:t/0","title":"Protean.Node.t/0","type":"type"},{"doc":"","ref":"Protean.ConfigError.html","title":"Protean.ConfigError","type":"exception"}]